{% load otree %}
{% for error in form.non_field_errors %}
    <div class="alert alert-error">
        <a class="close" href="#" data-dismiss="alert">×</a>
        {{ error }}
    </div><!--- .alert.alert-error -->
{% endfor %}

{% for field in form.visible_fields %}
    {% formfield field %}
{% endfor %}

{% if not form.visible_fields %}{% for field in form.hidden_fields %}{{ field }}{% endfor %}{% endif %}
