{% load static otree %}
{# spaces to make debug info less distracting #}
<br/>
<br/>
<br/>


<div class="card debug-info"> {# .debug-info so it's visible in inspector #}
    <h5 class="card-header">Debug info</h5>

    {% include 'otree/includes/hidden_form_errors.html' %}

{% for table in view.debug_tables|default:None %}
    <div class="card-body">
        <h6 class="card-title">{{ table.title }}</h6>
        <div class="scrollable">
            <table class="table table-hover table-debug">
                {% for k, v in table.rows %}
                    <tr>
                        <td class="debug-var-name">{{ k }}</td>
                        <td class="debug-var-value">
                            <code>{{ v }}</code>
                        </td>
                    </tr>
                {% endfor %}
            </table>
        </div>
    </div>
{% endfor %}

</div>
